from mlpfile import *


def array1d(x):
    return "{" + ", ".join([str(xi) for xi in x]) + "}"


def array2d(X):
    return "{\n" + ",\n".join([array1d(row) for row in X]) + "\n}"


def codegen_c(model, f):
    # static data
    for i, layer in enumerate(model.layers):
        if layer.type == LayerType.Linear:
            rows, cols = layer.W.shape
            f.write(f"float W_{i}[{rows}][{cols}] = {array2d(layer.W)};\n")
            f.write(f"float b_{i}[{rows}] = {array1d(layer.b)};\n")

    # workspace
    linears = [layer for layer in model.layers if layer.type == LayerType.Linear]
    workdim = max(
        [lay.W.shape[0] for lay in linears]
        + [lay.W.shape[1] for lay in linears]
    )
    f.write(f"float work[2][{workdim}];\n")

    # code
    size = model.input_dim()
    f.write(f"void forward(float const *x, float *y) {{")
    f.write(
f"""
for (int i = 0; i < {size}; ++i) {{
    work[0][i] = x[i];
}}
""")
    src = 0
    dst = 1
    for ilayer, layer in enumerate(model.layers):
        if layer.type == LayerType.Linear:
            newsize = layer.W.shape[0]
            f.write(
f"""
for (int i = 0; i < {newsize}; ++i) {{
    work[{dst}][i] = b_{ilayer}[i];
    for (int j = 0; j < {size}; ++j) {{
        work[{dst}][i] += W_{ilayer}[i][j] * work[{src}][j];
    }}
}}
""")
            size = newsize
        elif layer.type == LayerType.ReLU:
            # Avoid importing math.h just for fmaxf.
            f.write(
f"""
for (int i = 0; i < {size}; i++) {{
    work[{dst}][i] = (work[{src}][i] < 0.0f) ? 0.0f : work[{src}][i];
}}
""")
        else:
            raise ValueError("layer type:", layer.type)
        src, dst = dst, src
    f.write(
f"""
for (int i = 0; i < {size}; ++i) {{
    y[i] = work[{src}][i];
}}
}}
""")


def codegen_eigen(model, f):
    """Generates C++ code with C-only API to evaluate the model's forward pass.

    Takes ``mlpfile.Model`` and open file-like object. See the docstring of
    ``codegen_compile`` for details.
    """

    f.write("#include <Eigen/Dense>\n")

    # static data
    for i, layer in enumerate(model.layers):
        if layer.type == LayerType.Linear:
            rows, cols = layer.W.shape
            # I would have preferred to store data directly in Eigen::Matrices
            # instead of the extra complexity of these Eigen::Maps. But for
            # some reason the initializer list constructor of Eigen::Matrix
            # wasn't working even though I was using C++11.
            f.write(f"float arr_W_{i}[{rows}][{cols}] = {array2d(layer.W)};\n")
            f.write(f"float arr_b_{i}[{rows}] = {array1d(layer.b)};\n")
            f.write(f"Eigen::Map<Eigen::Matrix<float, {rows}, {cols}, Eigen::RowMajor> > W_{i} (arr_W_{i}[0]);\n")
            f.write(f"Eigen::Map<Eigen::Matrix<float, {rows}, 1> > b_{i} (arr_b_{i});\n")

    # workspace
    linears = [layer for layer in model.layers if layer.type == LayerType.Linear]
    workdim = max(
        [lay.W.shape[0] for lay in linears]
        + [lay.W.shape[1] for lay in linears]
    )
    f.write(f"Eigen::Matrix<float, {workdim}, 1> work[2];\n")

    # code
    size = model.input_dim()
    f.write(f'extern "C" void forward(float const *xptr, float *yptr) {{\n')
    f.write(f"Eigen::Map<Eigen::Matrix<float, {size}, 1> const> x(xptr);\n")
    f.write(f"work[0].head<{size}>() = x;\n")
    src = 0
    dst = 1
    for ilayer, layer in enumerate(model.layers):
        if layer.type == LayerType.Linear:
            newsize = layer.W.shape[0]
            f.write(f"work[{dst}].head<{newsize}>() = W_{ilayer} * work[{src}].head<{size}>() + b_{ilayer};\n")
            size = newsize
        elif layer.type == LayerType.ReLU:
            f.write(f"work[{dst}].head<{size}>() = work[{src}].head<{size}>().array().max(0);\n")
        else:
            raise ValueError("layer type:", layer.type)
        src, dst = dst, src
    f.write(f"Eigen::Map<Eigen::Matrix<float, {model.output_dim()}, 1> > y(yptr);\n")
    f.write(f"y = work[{src}].head<{size}>();\n")
    f.write("}")



import torch


def mlp(indim, outdim, hidden):
    """Construct a multilayer perceptron that can be represented by mlpfile.

    Will use ReLU activation automatically.

    PyTorch contains some tools that modify the training of an MLP but do not
    change its architecture, such as spectral normalization. To use these
    tools, please construct the ``torch.nn.Sequential`` yourself instead of using
    this function.

    Args:
        indim (int): Input dimension.
        outdim (int): Output dimension.
        hidden (List[int]): List of hidden layer dimensions.

    Returns:
        a ``torch.nn.Sequential`` module.
    """
    dims = [indim] + hidden
    layers = []
    for i in range(len(dims) - 1):
        layers.append(torch.nn.Linear(dims[i], dims[i + 1]))
        layers.append(torch.nn.ReLU())
    layers.append(torch.nn.Linear(dims[-1], outdim))
    return torch.nn.Sequential(*layers)


def write(model: torch.nn.Sequential, path):
    """Writes a fully connected ReLU network to our file format.

    See top-level module documentation for binary format information.

    Args:
        model (torch.nn.Sequential): Trained model, such as one with the
            structure returned by :meth:`mlp()`.
        path (path-like): A `path-like object
            <https://docs.python.org/3/glossary.html#term-path-like-object>`_
            where the file should be written.
    """
    LINEAR = 2
    RELU = 3
    with open(path, "wb") as f:
        def u32(i):
            f.write(i.to_bytes(4, byteorder="little", signed=False))
        u32(len(model))

        # First pass - get input size.
        size = 0
        for m in model:
            if isinstance(m, torch.nn.Linear):
                size = m.weight.shape[1]
                break
        if size == 0:
            raise ValueError("Cannot deduce input size.")
        u32(size)

        # Second pass - metadata.
        for m in model:
            if isinstance(m, torch.nn.Linear):
                assert size != 0
                osize, isize = m.weight.shape
                assert osize != 0
                assert isize == size
                u32(LINEAR)
                u32(osize)
                size = osize
            elif isinstance(m, torch.nn.ReLU):
                u32(RELU)
            else:
                raise ValueError

        # Third pass - weight data.
        # TODO: We should enforce little-endian explicitly, here + the loader.
        for m in model:
            if isinstance(m, torch.nn.Linear):
                W = m.weight.detach().to(torch.float32).numpy()
                W = W.reshape(*W.shape, order="C")
                bW = W.tobytes()
                assert len(bW) == W.size * 4
                f.write(bW)
                b = m.bias.detach().to(torch.float32).numpy()
                bb = b.tobytes()
                assert len(bb) == b.size * 4
                f.write(bb)



from copy import deepcopy
import ctypes
import tempfile

import numpy as np
import pytest
import torch

import mlpfile
import mlpfile.torch


INDIM = 40
OUTDIM = 10
HIDDEN = [100] * 2
NET = mlpfile.torch.mlp(INDIM, OUTDIM, HIDDEN)
NET_NOTEVAL = deepcopy(NET)
NET.eval()
JAC = torch.func.jacrev(NET)


def norm2(x):
    return np.sum(x ** 2)


def softmax(x):
    e = np.exp(x)
    return e / np.sum(e)


# The point of this is to make sure we write the file exactly once and then
# clean up after all tests are complete. See the pytest "How to use fixtures"
# docs for more info.
@pytest.fixture(scope="module", autouse=True)
def model():
    with tempfile.NamedTemporaryFile() as f:
        mlpfile.torch.write(NET, f.name)
        model = mlpfile.Model.load(f.name)
        yield model


def test_forward(model):
    for _ in range(100):
        x = torch.randn(INDIM, requires_grad=False)
        y = model.forward(x)
        ytorch = NET.forward(x).detach().numpy()
        # TODO: Why do we need to loosen the tolerance?
        assert np.all(np.isclose(ytorch, y, atol=1e-6, rtol=1e-4))


def test_jacobian(model):
    for i in range(100):
        x = torch.randn(INDIM, requires_grad=False)
        J = model.jacobian(x)
        Jtorch = JAC(x).detach().numpy()
        # TODO: Why do we need to loosen the tolerance?
        assert np.all(np.isclose(J.flat, Jtorch.flat, atol=1e-6, rtol=1e-4))


def test_jacobian_params(model):
    assert len(model.layers) == len(NET_NOTEVAL)
    for i in range(100):
        x = torch.randn(INDIM, requires_grad=False)
        J = model.jacobian_params(x)
        for j in range(OUTDIM):
            NET_NOTEVAL.zero_grad()
            y = NET_NOTEVAL.forward(x)
            y[j].backward()
            for layer, Jlayer, tparam in zip(model.layers, J, NET_NOTEVAL):
                if layer.type != mlpfile.LayerType.Linear:
                    continue
                d_yj_W = Jlayer.dW[j].reshape(layer.W.shape)
                # TODO: Why do we need to loosen the tolerance?
                assert np.allclose(d_yj_W, tparam.weight.grad, atol=1e-6, rtol=1e-4)
                assert np.allclose(Jlayer.db[j], tparam.bias.grad, atol=1e-6, rtol=1e-4)


class GradCaseSqErr:
    def __init__(self):
        self.randomize()
        self.lossgrad = mlpfile.squared_error
        self.rate = 1e-2

    def randomize(self):
        self.ytrue = np.random.normal(size=OUTDIM)

    def loss(self, y):
        # Accounting for the 1/2 is important!
        return 0.5 * norm2(self.ytrue - y)

    def opt_loss(self):
        return 0.0


# In earlier prototype we set ytrue to a uniform (wrt Borel measure) sample
# from the simplex. However, it needed unreasonably many iterations to reach
# low error in test_ogd_onepoint. Choosing these sharp distributions (like we
# actually use in supervised classification) and a high learning rate gives
# fast convergence. TODO: Build more intuition for softmax cross-entropy to
# understand this. Why would fitting a high-entropy distribution require more
# steps?
class GradCaseXEnt:
    def __init__(self):
        self.randomize()
        self.lossgrad = mlpfile.softmax_cross_entropy
        self.rate = 1e-1

    def randomize(self):
        self.ytrue = (0.1 / (OUTDIM - 1)) * np.ones(OUTDIM)
        hot = np.random.choice(OUTDIM)
        self.ytrue[hot] = 0.9

    def loss(self, y):
        s = np.exp(y)
        s /= np.sum(s)
        return -np.sum(self.ytrue * np.log(s))

    def opt_loss(self):
        return -np.sum(np.log(self.ytrue) * self.ytrue)


GRAD_CASES = [GradCaseSqErr(), GradCaseXEnt()]
GRAD_IDS = ["squared-error", "cross-entropy"]


@pytest.mark.parametrize("gradcase", GRAD_CASES, ids=GRAD_IDS)
def test_ogd_onepoint(model, gradcase):
    x = np.random.normal(size=INDIM)
    # Full log makes debugging easier - plots, etc.
    trials = 100
    iters = 100
    gamma = 1.0 ** (1.0 / iters)
    rate = gradcase.rate
    for trial in range(trials):
        model2 = deepcopy(model)
        gradcase.randomize()
        print("initial:", gradcase.loss(model2.forward(x)))
        print(softmax(model2.forward(x)))
        print("opt:", gradcase.opt_loss())
        print(gradcase.ytrue)
        errs = []
        for i in range(iters):
            y = model2.forward(x);
            errs.append(gradcase.loss(y))
            model2.grad_update(x, gradcase.ytrue, gradcase.lossgrad, rate)
            rate *= gamma
        errs.append(gradcase.loss(model2.forward(x)))
        print("final:", softmax(model2.forward(x)))
        print()
        # We should be able to fit perfectly using last layer's bias.
        assert errs[-1] < gradcase.opt_loss() + 5e-5


@pytest.mark.parametrize("gradcase", GRAD_CASES, ids=GRAD_IDS)
def test_ogd_finitediff(model, gradcase):
    x = np.random.normal(size=INDIM)
    loss_original = gradcase.loss(model.forward(x))
    rate = gradcase.rate / 10

    trials = 100
    for trial in range(trials):
        model2 = deepcopy(model)
        model2.grad_update(x, gradcase.ytrue, gradcase.lossgrad, rate)

        # Reverse engineer what the gradient was.
        layers = model.layers
        grad_dot_step = 0.0
        for i in range(len(layers)):
            if layers[i].type == mlpfile.LayerType.Linear:
                gradW = (model2.layers[i].W - model.layers[i].W) / rate
                grad_dot_step -= rate * norm2(gradW.flatten())
                gradb = (model2.layers[i].b - model.layers[i].b) / rate
                grad_dot_step -= rate * norm2(gradb)

        # Predict what the new loss should be using first-order approximation.
        loss_predicted = loss_original + grad_dot_step
        assert loss_predicted < loss_original
        loss_actual = gradcase.loss(model2.forward(x))
        print(
            f"loss {gradcase.lossgrad.__name__}:\n"
            f"original: {loss_original}, "
            f"predicted: {loss_predicted}, "
            f"actual: {loss_actual}"
        )
        assert np.abs(loss_predicted - loss_actual) / loss_original < 5e-3


def test_ogd_multipoint(model):
    model = deepcopy(model)
    rate = 1e-2
    # Overfit a few random data points.
    # See "Understanding Deep Learning Requires Rethinking Generalization"
    N = 20
    xs = np.random.normal(size=(N, INDIM))
    ys = np.random.normal(size=(N, OUTDIM))
    # Full log makes debugging easier - plots, etc.
    errs = []
    for epoch in range(50):
        for i in range(N):
            y = model.forward(xs[i]);
            model.grad_update(xs[i], ys[i], mlpfile.squared_error, rate)
        errs.append(np.mean([
            norm2(model.forward(x) - y)
            for x, y in zip(xs, ys)
        ]))
    assert errs[0] > 1
    assert errs[-1] < 1e-2


def test_random():
    r = mlpfile.Model.random(2, [3, 4], 5)
    assert len(r.layers) == 5
    types = [lay.type for lay in r.layers]
    linear = mlpfile.LayerType.Linear
    relu = mlpfile.LayerType.ReLU
    assert types == [linear, relu, linear, relu, linear]
    y = r.forward([0, 0])
    assert y.size == 5
    r.jacobian([0, 0])


# Basic test of dimensions.
def test_jacobian_params_dim():
    model = mlpfile.Model.random(INDIM, HIDDEN, OUTDIM)
    x = np.random.normal(size=INDIM)

    jac_params = model.jacobian_params(x)
    assert len(jac_params) == len(model.layers)

    outdim = model.output_dim()

    for i, layer_jac in enumerate(jac_params):
        layer = model.layers[i]
        if layer.type == mlpfile.LayerType.Linear:
            assert layer_jac.dW.shape == (outdim, np.prod(layer.W.shape))
            assert layer_jac.db.shape == (outdim, layer.b.size)
        else:
            # ReLU layers should have empty Jacobians.
            assert layer_jac.dW.shape == (0, 0)
            assert layer_jac.db.shape == (0, 0)


def test_jacobian_params_analytic():

    model = mlpfile.Model()

    layer0 = mlpfile.Layer()
    layer0.type = mlpfile.LayerType.Linear
    layer0.W = np.eye(2)
    layer0.b = np.zeros(2)

    layer1 = mlpfile.Layer()
    layer1.type = mlpfile.LayerType.ReLU

    layer2 = mlpfile.Layer()
    layer2.type = mlpfile.LayerType.Linear
    layer2.W = np.ones((1, 2))
    layer2.b = np.zeros(1)

    model.layers = [layer0, layer1, layer2]

    # no ReLU
    x = np.array([1, 2])

    j0, _, j2 = model.jacobian_params(x)
    assert j0.dW.shape == (1, 4)
    assert np.all(j0.dW.reshape((2, 2)) == [x, x])
    assert np.all(j0.db == [[1, 1]])

    assert np.all(j2.dW == x[None, :])
    assert np.all(j2.db == [[1]])

    # ReLU
    x = np.array([-1, 1])
    j0, _, j2 = model.jacobian_params(x)
    assert j0.dW.shape == (1, 4)
    assert np.all(j0.dW.reshape((2, 2)) == [0 * x, x])
    assert np.all(j0.db == [[0, 1]])

    assert np.all(j2.dW == [[0, 1]])
    assert np.all(j2.db == [[1]])


def test_jacobian_params_finitediff():

    trials = 10
    for trial in range(trials):
        x = np.random.normal(size=2)
        # use prime numbers to bring out anything related to transpose
        model = mlpfile.Model.random(2, [3], 5)
        for layer in model.layers:
            layer.W = np.abs(layer.W)
        y = model.forward(x)
        J = model.jacobian_params(x)

        EPS = 1e-4
        for _ in range(10):
            model2 = deepcopy(model)
            dy_pred = np.zeros(5)
            for layer, Jlayer in zip(model2.layers, J):
                if layer.type == mlpfile.LayerType.Linear:
                    dW = EPS * np.random.normal(size=layer.W.shape)
                    db = EPS * np.random.normal(size=layer.b.shape)
                    # pybind11 doesn't give us +=. Not sure if worth fixing.
                    layer.W = layer.W + dW
                    layer.b = layer.b + db
                    dy_pred += Jlayer.dW @ dW.flat / EPS
                    dy_pred += Jlayer.db @ db / EPS
            dy_actual = (model2.forward(x) - y) / EPS
            # atol has to be pretty loose because ReLU is not smooth.
            assert np.allclose(dy_pred, dy_actual, atol=5e-2)


def test_specnorm_basic():
    model = mlpfile.Model()

    layer = mlpfile.Layer()
    layer.type = mlpfile.LayerType.Linear
    layer.W = np.diag([1, 2])
    layer.b = np.zeros(2)

    model.layers = [layer]
    model.spec_norm_init()
    # nonsense checks
    model.spec_norm_update(power_iterations=-1)
    model.spec_norm_update(power_iterations=0)
    # check u
    model.spec_norm_update(power_iterations=100)
    u = model.spec_norm[0]
    target = [0, 1]
    assert np.allclose(u, target) or np.allclose(-u, target)
    # check W
    L = np.linalg.norm(model.layers[0].W, ord=2)
    assert np.isclose(L, 1)


def test_specnorm_doesnt_magnify():
    model = mlpfile.Model()

    layer = mlpfile.Layer()
    layer.type = mlpfile.LayerType.Linear
    layer.W = np.diag([0.1])
    layer.b = np.zeros(1)
    model.layers = [layer]

    model.spec_norm_init()
    model.spec_norm_update(power_iterations=100)
    assert np.isclose(model.layers[0].W[0, 0], 0.1)


def test_specnorm_fighting_ogd():
    # Regression to map 1 to 10 and -1 to -10. Interleave steps of OGD and
    # spec_norm_update. Make sure W stays at 1 even though 10 would be optimal.
    model = mlpfile.Model()

    layer = mlpfile.Layer()
    layer.type = mlpfile.LayerType.Linear
    layer.W = np.diag([1])
    layer.b = np.zeros(1)
    model.layers = [layer]

    model.spec_norm_init()

    xs = [np.array([1]), np.array([-1])]
    ys = [np.array([10]), np.array([-10])]

    rate = 1e-1
    iters = 100

    for i in range(iters):
        for x, y in zip(xs, ys):
            model.grad_update(x, y, mlpfile.squared_error, rate)
        model.spec_norm_update(power_iterations=1)
    assert np.allclose(model.layers[0].W[0, 0], 1.0)

    # Without SpecNorm
    for i in range(iters):
        for x, y in zip(xs, ys):
            model.grad_update(x, y, mlpfile.squared_error, rate)
    assert np.allclose(model.layers[0].W[0, 0], 10.0)


def test_cpp_dir(capfd):
    mlpfile.cpp_dir()
    out, _ = capfd.readouterr()
    assert out.endswith("cpp")


@pytest.mark.parametrize("eigen", [True, False])
def test_codegen(model, eigen):
    lib = mlpfile.codegen(model, eigen=eigen, compile=True)
    x = np.random.normal(size=INDIM).astype(np.float32)
    y = np.zeros(OUTDIM, dtype=np.float32)
    xptr = x.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
    yptr = y.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
    lib.forward(xptr, yptr)
    y2 = model.forward(x)
    assert np.allclose(y, y2, atol=1e-6)



"""Submodule of ``mlpfile`` for functions that must import PyTorch.

Note that this submodule is written in Python. None of this functionality is available from the ``mlpfile`` C++ API.
"""

from .torch import mlp, write

__all__ = ["mlp", "write"]



"""mlpfile: Multilayer perceptron file format and evaluation."""

import ctypes
import os
import subprocess
import tempfile
import warnings

from _mlpfile import *
from ._codegen import codegen_c, codegen_eigen


__all__ = [
    "Model", "Layer", "LayerType", "LayerJacobian", # Bound class/enum.
    "squared_error", "softmax_cross_entropy", # Bound function.
    "codegen", "cpp_dir", # Pure Python.
]


def cpp_dir():
    """Prints the directory containing the C++ header and source file.

    If you install the package via ``pip``, you can use the files in your build
    system. For example, a Makefile might contain::

        CXXFLAGS := -I$(shell python -c "__import__('mlpfile').cpp_dir()")
    """
    import os.path
    print(os.path.dirname(__file__) + "/cpp", end="")  # no newline


def codegen(model, outdir=None, eigen=True, compile=False):
    """Generates and optionally compiles C/C++ code to evaluate the model's
    forward pass.

    The generated code has the function::

        void forward(float const *x, float *y)

    Note that size arguments are not needed because they are known at compile time.

    If ``compile=True``, also compiles the code into a shared library and loads
    that library with ``ctypes``. Calling from Python+NumPy looks like::

        lib = mlpfile.codegen(model, compile=True)
        x = <my input of length model.input_dim()>.astype(np.float32)
        y = np.zeros(model.output_dim(), dtype=np.float32)
        xptr = x.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
        yptr = y.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
        lib.forward(xptr, yptr)
        # Now y has the output.

    Even if you don't want to use the compiled code from Python,
    ``compile=True`` can be helpful because it knows the correct compiler
    flags. This is especially important when ``eigen=True``.

    To compile the Eigen code yourself, these flags ensure the code works in a
    malloc-free (embedded/realtime) environment and can be linked as C::

        -fno-exceptions -DEIGEN_NO_MALLOC

    Args:
        model (Model): Model.
        outdir (str, optional): If specified, writes header and code to this
            directory. If ``compile=True``, also stores object file and shared
            library. If ``outdir=None``, uses a temporary directory deleted
            immediately after call. Note: ``outdir=None`` only makes sense when
            ``compile=True``.
        eigen (bool): Use Eigen internally. If true, we generate code that uses
            the Eigen C++ library. Eigen's matrix-vector multiply is much
            faster than our hand-written naive one when ``eigen=False``.
            Generates a function with ``extern "C"`` linkage, no dynamic
            memory, and no exceptions. Therefore, the compiled object/library
            can be used in C projects, with a C linker, even though it was
            compiled by a C++ compiler.
        compile (bool): Compiles the code into a shared library and loads it
            with ``ctypes``. If ``outdir=None``, the returned ``ctypes.CDLL``
            will be loaded before the temporary dir is deleted, so you can
            still use the library for the remainder of the Python process.

    Returns:
        If ``compile=True``, returns a ``ctypes.CDLL`` library with function
        ``forward`` described in overview. Otherwise, returns nothing.
    """
    if not compile and outdir is None:
        warnings.warn(
            "Will have no observable effect (besides errors)"
            " with compile=False and outdir=None."
        )

    if outdir is None:
        with tempfile.TemporaryDirectory() as d:
            return codegen(model, outdir=d, eigen=eigen, compile=compile)
    else:
        os.makedirs(outdir, exist_ok=True)
        print("writing to", outdir)

    header = outdir + "/mlp.h"
    src = outdir + "/mlp." + ("cpp" if eigen else "c")
    obj = outdir + "/mlp.o"
    lib = outdir + "/mlp.so"

    with open(header, "w") as f:
        f.write(f"#pragma once\n\n")
        f.write(f"#define MLP_INPUT_DIM {model.input_dim()}\n")
        f.write(f"#define MLP_OUTPUT_DIM {model.output_dim()}\n\n")
        f.write("void forward(float const *x, float *y);\n")

    with open(src, "w") as f:
        if eigen:
            codegen_eigen(model, f)
        else:
            codegen_c(model, f)

    if not compile:
        return

    flags_both = ["-c", "-fPIC", "-O3", "-o", obj]
    if eigen:
        # TODO: Useful to allow getting Eigen somewhere else besides eigenpip?
        import eigenpip
        incl = "-I" + eigenpip.get_include()
        flags_eig = [incl, "-fno-exceptions", "-DEIGEN_NO_MALLOC"]
        result_compile = subprocess.run(["c++"] + flags_both + flags_eig + [src])
    else:
        result_compile = subprocess.run(["cc"] + flags_both + [src])
    assert result_compile.returncode == 0
    result_shared = subprocess.run(["cc", "-shared", "-o", lib, obj])
    assert result_shared.returncode == 0
    return ctypes.cdll.LoadLibrary(lib)



# Available at setup time due to pyproject.toml
import eigenpip
from pybind11.setup_helpers import Pybind11Extension
from setuptools import setup

__version__ = "0.5.0"

ext_modules = [
    Pybind11Extension("_mlpfile",
        ["mlpfile/cpp/bindings.cpp", "mlpfile/cpp/mlpfile.cpp"],
        include_dirs = [eigenpip.get_include()],
        cxx_std=11,
        ),
]

setup(
    name="mlpfile",
    version=__version__,
    author="James A. Preiss",
    author_email="jamesalanpreiss@gmail.com",
    url="https://github.com/jpreiss/mlpfile",
    description="Multilayer perceptron file format and evaluation.",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    ext_modules=ext_modules,
    packages=["mlpfile", "mlpfile.torch"],
    package_data={
        "mlpfile": ["cpp/mlpfile.h", "cpp/mlpfile.cpp"]
    },
    requires=["numpy"],
    extras_require={
        "test": ["pytest", "torch"],
        "torch": "torch",
    },
    zip_safe=False,  # TODO: Understand.
    python_requires=">=3.7",
)
